from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ManagedJobStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    MANAGED_JOB_STATUS_UNSPECIFIED: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_PENDING: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_SUBMITTED: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_STARTING: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_RUNNING: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_RECOVERING: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_CANCELLING: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_SUCCEEDED: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_CANCELLED: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_FAILED: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_FAILED_SETUP: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_FAILED_PRECHECKS: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_FAILED_NO_RESOURCE: _ClassVar[ManagedJobStatus]
    MANAGED_JOB_STATUS_FAILED_CONTROLLER: _ClassVar[ManagedJobStatus]

class ManagedJobScheduleState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    MANAGED_JOB_SCHEDULE_STATE_UNSPECIFIED: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_INVALID: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_INACTIVE: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_WAITING: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_ALIVE_WAITING: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_LAUNCHING: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_ALIVE_BACKOFF: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_ALIVE: _ClassVar[ManagedJobScheduleState]
    MANAGED_JOB_SCHEDULE_STATE_DONE: _ClassVar[ManagedJobScheduleState]
MANAGED_JOB_STATUS_UNSPECIFIED: ManagedJobStatus
MANAGED_JOB_STATUS_PENDING: ManagedJobStatus
MANAGED_JOB_STATUS_SUBMITTED: ManagedJobStatus
MANAGED_JOB_STATUS_STARTING: ManagedJobStatus
MANAGED_JOB_STATUS_RUNNING: ManagedJobStatus
MANAGED_JOB_STATUS_RECOVERING: ManagedJobStatus
MANAGED_JOB_STATUS_CANCELLING: ManagedJobStatus
MANAGED_JOB_STATUS_SUCCEEDED: ManagedJobStatus
MANAGED_JOB_STATUS_CANCELLED: ManagedJobStatus
MANAGED_JOB_STATUS_FAILED: ManagedJobStatus
MANAGED_JOB_STATUS_FAILED_SETUP: ManagedJobStatus
MANAGED_JOB_STATUS_FAILED_PRECHECKS: ManagedJobStatus
MANAGED_JOB_STATUS_FAILED_NO_RESOURCE: ManagedJobStatus
MANAGED_JOB_STATUS_FAILED_CONTROLLER: ManagedJobStatus
MANAGED_JOB_SCHEDULE_STATE_UNSPECIFIED: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_INVALID: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_INACTIVE: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_WAITING: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_ALIVE_WAITING: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_LAUNCHING: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_ALIVE_BACKOFF: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_ALIVE: ManagedJobScheduleState
MANAGED_JOB_SCHEDULE_STATE_DONE: ManagedJobScheduleState

class JobIds(_message.Message):
    __slots__ = ("ids",)
    IDS_FIELD_NUMBER: _ClassVar[int]
    ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, ids: _Optional[_Iterable[int]] = ...) -> None: ...

class UserHashes(_message.Message):
    __slots__ = ("hashes",)
    HASHES_FIELD_NUMBER: _ClassVar[int]
    hashes: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, hashes: _Optional[_Iterable[str]] = ...) -> None: ...

class Statuses(_message.Message):
    __slots__ = ("statuses",)
    STATUSES_FIELD_NUMBER: _ClassVar[int]
    statuses: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, statuses: _Optional[_Iterable[str]] = ...) -> None: ...

class GetVersionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetVersionResponse(_message.Message):
    __slots__ = ("controller_version",)
    CONTROLLER_VERSION_FIELD_NUMBER: _ClassVar[int]
    controller_version: str
    def __init__(self, controller_version: _Optional[str] = ...) -> None: ...

class GetJobTableRequest(_message.Message):
    __slots__ = ("skip_finished", "accessible_workspaces", "job_ids", "workspace_match", "name_match", "pool_match", "page", "limit", "user_hashes", "statuses", "show_jobs_without_user_hash")
    SKIP_FINISHED_FIELD_NUMBER: _ClassVar[int]
    ACCESSIBLE_WORKSPACES_FIELD_NUMBER: _ClassVar[int]
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_MATCH_FIELD_NUMBER: _ClassVar[int]
    NAME_MATCH_FIELD_NUMBER: _ClassVar[int]
    POOL_MATCH_FIELD_NUMBER: _ClassVar[int]
    PAGE_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    USER_HASHES_FIELD_NUMBER: _ClassVar[int]
    STATUSES_FIELD_NUMBER: _ClassVar[int]
    SHOW_JOBS_WITHOUT_USER_HASH_FIELD_NUMBER: _ClassVar[int]
    skip_finished: bool
    accessible_workspaces: _containers.RepeatedScalarFieldContainer[str]
    job_ids: JobIds
    workspace_match: str
    name_match: str
    pool_match: str
    page: int
    limit: int
    user_hashes: UserHashes
    statuses: Statuses
    show_jobs_without_user_hash: bool
    def __init__(self, skip_finished: bool = ..., accessible_workspaces: _Optional[_Iterable[str]] = ..., job_ids: _Optional[_Union[JobIds, _Mapping]] = ..., workspace_match: _Optional[str] = ..., name_match: _Optional[str] = ..., pool_match: _Optional[str] = ..., page: _Optional[int] = ..., limit: _Optional[int] = ..., user_hashes: _Optional[_Union[UserHashes, _Mapping]] = ..., statuses: _Optional[_Union[Statuses, _Mapping]] = ..., show_jobs_without_user_hash: bool = ...) -> None: ...

class ManagedJobInfo(_message.Message):
    __slots__ = ("job_id", "task_id", "job_name", "task_name", "job_duration", "workspace", "status", "schedule_state", "resources", "cluster_resources", "cluster_resources_full", "cloud", "region", "infra", "accelerators", "recovery_count", "details", "failure_reason", "user_name", "user_hash", "submitted_at", "start_at", "end_at", "user_yaml", "entrypoint", "metadata", "pool", "pool_hash")
    class AcceleratorsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    TASK_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_DURATION_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    SCHEDULE_STATE_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_RESOURCES_FULL_FIELD_NUMBER: _ClassVar[int]
    CLOUD_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    INFRA_FIELD_NUMBER: _ClassVar[int]
    ACCELERATORS_FIELD_NUMBER: _ClassVar[int]
    RECOVERY_COUNT_FIELD_NUMBER: _ClassVar[int]
    DETAILS_FIELD_NUMBER: _ClassVar[int]
    FAILURE_REASON_FIELD_NUMBER: _ClassVar[int]
    USER_NAME_FIELD_NUMBER: _ClassVar[int]
    USER_HASH_FIELD_NUMBER: _ClassVar[int]
    SUBMITTED_AT_FIELD_NUMBER: _ClassVar[int]
    START_AT_FIELD_NUMBER: _ClassVar[int]
    END_AT_FIELD_NUMBER: _ClassVar[int]
    USER_YAML_FIELD_NUMBER: _ClassVar[int]
    ENTRYPOINT_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    POOL_HASH_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    task_id: int
    job_name: str
    task_name: str
    job_duration: float
    workspace: str
    status: ManagedJobStatus
    schedule_state: ManagedJobScheduleState
    resources: str
    cluster_resources: str
    cluster_resources_full: str
    cloud: str
    region: str
    infra: str
    accelerators: _containers.ScalarMap[str, float]
    recovery_count: int
    details: str
    failure_reason: str
    user_name: str
    user_hash: str
    submitted_at: float
    start_at: float
    end_at: float
    user_yaml: str
    entrypoint: str
    metadata: _containers.ScalarMap[str, str]
    pool: str
    pool_hash: str
    def __init__(self, job_id: _Optional[int] = ..., task_id: _Optional[int] = ..., job_name: _Optional[str] = ..., task_name: _Optional[str] = ..., job_duration: _Optional[float] = ..., workspace: _Optional[str] = ..., status: _Optional[_Union[ManagedJobStatus, str]] = ..., schedule_state: _Optional[_Union[ManagedJobScheduleState, str]] = ..., resources: _Optional[str] = ..., cluster_resources: _Optional[str] = ..., cluster_resources_full: _Optional[str] = ..., cloud: _Optional[str] = ..., region: _Optional[str] = ..., infra: _Optional[str] = ..., accelerators: _Optional[_Mapping[str, float]] = ..., recovery_count: _Optional[int] = ..., details: _Optional[str] = ..., failure_reason: _Optional[str] = ..., user_name: _Optional[str] = ..., user_hash: _Optional[str] = ..., submitted_at: _Optional[float] = ..., start_at: _Optional[float] = ..., end_at: _Optional[float] = ..., user_yaml: _Optional[str] = ..., entrypoint: _Optional[str] = ..., metadata: _Optional[_Mapping[str, str]] = ..., pool: _Optional[str] = ..., pool_hash: _Optional[str] = ...) -> None: ...

class GetJobTableResponse(_message.Message):
    __slots__ = ("jobs", "total", "total_no_filter", "status_counts")
    class StatusCountsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    JOBS_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    TOTAL_NO_FILTER_FIELD_NUMBER: _ClassVar[int]
    STATUS_COUNTS_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[ManagedJobInfo]
    total: int
    total_no_filter: int
    status_counts: _containers.ScalarMap[str, int]
    def __init__(self, jobs: _Optional[_Iterable[_Union[ManagedJobInfo, _Mapping]]] = ..., total: _Optional[int] = ..., total_no_filter: _Optional[int] = ..., status_counts: _Optional[_Mapping[str, int]] = ...) -> None: ...

class GetAllJobIdsByNameRequest(_message.Message):
    __slots__ = ("job_name",)
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    job_name: str
    def __init__(self, job_name: _Optional[str] = ...) -> None: ...

class GetAllJobIdsByNameResponse(_message.Message):
    __slots__ = ("job_ids",)
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    job_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, job_ids: _Optional[_Iterable[int]] = ...) -> None: ...

class CancelJobsRequest(_message.Message):
    __slots__ = ("current_workspace", "user_hash", "all_users", "job_ids", "job_name", "pool_name")
    CURRENT_WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    USER_HASH_FIELD_NUMBER: _ClassVar[int]
    ALL_USERS_FIELD_NUMBER: _ClassVar[int]
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    POOL_NAME_FIELD_NUMBER: _ClassVar[int]
    current_workspace: str
    user_hash: str
    all_users: bool
    job_ids: JobIds
    job_name: str
    pool_name: str
    def __init__(self, current_workspace: _Optional[str] = ..., user_hash: _Optional[str] = ..., all_users: bool = ..., job_ids: _Optional[_Union[JobIds, _Mapping]] = ..., job_name: _Optional[str] = ..., pool_name: _Optional[str] = ...) -> None: ...

class CancelJobsResponse(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class StreamLogsRequest(_message.Message):
    __slots__ = ("job_name", "job_id", "follow", "controller", "tail")
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    FOLLOW_FIELD_NUMBER: _ClassVar[int]
    CONTROLLER_FIELD_NUMBER: _ClassVar[int]
    TAIL_FIELD_NUMBER: _ClassVar[int]
    job_name: str
    job_id: int
    follow: bool
    controller: bool
    tail: int
    def __init__(self, job_name: _Optional[str] = ..., job_id: _Optional[int] = ..., follow: bool = ..., controller: bool = ..., tail: _Optional[int] = ...) -> None: ...

class StreamLogsResponse(_message.Message):
    __slots__ = ("log_line", "exit_code")
    LOG_LINE_FIELD_NUMBER: _ClassVar[int]
    EXIT_CODE_FIELD_NUMBER: _ClassVar[int]
    log_line: str
    exit_code: int
    def __init__(self, log_line: _Optional[str] = ..., exit_code: _Optional[int] = ...) -> None: ...



"""This module contains schemas used to validate objects.

Schemas conform to the JSON Schema specification as defined at
https://json-schema.org/
"""
import enum
from typing import Any, Dict, List, Tuple

from sky.skylet import autostop_lib
from sky.skylet import constants
from sky.utils import kubernetes_enums


def _check_not_both_fields_present(field1: str, field2: str):
    return {
        'oneOf': [{
            'required': [field1],
            'not': {
                'required': [field2]
            }
        }, {
            'required': [field2],
            'not': {
                'required': [field1]
            }
        }, {
            'not': {
                'anyOf': [{
                    'required': [field1]
                }, {
                    'required': [field2]
                }]
            }
        }]
    }


_AUTOSTOP_SCHEMA = {
    'anyOf': [
        {
            # Use boolean to disable autostop completely, e.g.
            #   autostop: false
            'type': 'boolean',
        },
        {
            # Shorthand to set idle_minutes by directly specifying, e.g.
            #   autostop: 5
            'anyOf': [{
                'type': 'string',
                'pattern': constants.TIME_PATTERN,
                'minimum': 0,
            }, {
                'type': 'integer',
            }]
        },
        {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                # TODO(luca): update field to use time units as well.
                'idle_minutes': {
                    'type': 'integer',
                    'minimum': 0,
                },
                'down': {
                    'type': 'boolean',
                },
                'wait_for': {
                    'type': 'string',
                    'case_insensitive_enum':
                        autostop_lib.AutostopWaitFor.supported_modes(),
                }
            },
        },
    ],
}


# Note: This is similar to _get_infra_pattern()
# but without the wildcard patterns.
def _get_volume_infra_pattern():
    # Building the regex pattern for the infra field
    # Format: cloud[/region[/zone]] or wildcards or kubernetes context
    # Match any cloud name (case insensitive)
    all_clouds = list(constants.ALL_CLOUDS)
    all_clouds.remove('kubernetes')
    cloud_pattern = f'(?i:({"|".join(all_clouds)}))'

    # Optional /region followed by optional /zone
    # /[^/]+ matches a slash followed by any characters except slash (region or
    # zone name)
    # The outer (?:...)? makes the entire region/zone part optional
    region_zone_pattern = '(?:/[^/]+(?:/[^/]+)?)?'

    # Kubernetes specific pattern - matches:
    # 1. Just the word "kubernetes" or "k8s" by itself
    # 2. "k8s/" or "kubernetes/" followed by any context name (which may contain
    # slashes)
    kubernetes_pattern = '(?i:kubernetes|k8s)(?:/.+)?'

    # Combine all patterns with alternation (|)
    # ^ marks start of string, $ marks end of string
    infra_pattern = (f'^(?:{cloud_pattern}{region_zone_pattern}|'
                     f'{kubernetes_pattern})$')
    return infra_pattern


def _get_infra_pattern():
    # Building the regex pattern for the infra field
    # Format: cloud[/region[/zone]] or wildcards or kubernetes context
    # Match any cloud name (case insensitive)
    all_clouds = list(constants.ALL_CLOUDS)
    all_clouds.remove('kubernetes')
    cloud_pattern = f'(?i:({"|".join(all_clouds)}))'

    # Optional /region followed by optional /zone
    # /[^/]+ matches a slash followed by any characters except slash (region or
    # zone name)
    # The outer (?:...)? makes the entire region/zone part optional
    region_zone_pattern = '(?:/[^/]+(?:/[^/]+)?)?'

    # Wildcard patterns:
    # 1. * - any cloud
    # 2. */region - any cloud with specific region
    # 3. */*/zone - any cloud, any region, specific zone
    wildcard_cloud = '\\*'  # Wildcard for cloud
    wildcard_with_region = '(?:/[^/]+(?:/[^/]+)?)?'

    # Kubernetes specific pattern - matches:
    # 1. Just the word "kubernetes" or "k8s" by itself
    # 2. "k8s/" or "kubernetes/" followed by any context name (which may contain
    # slashes)
    kubernetes_pattern = '(?i:kubernetes|k8s)(?:/.+)?'

    # Combine all patterns with alternation (|)
    # ^ marks start of string, $ marks end of string
    infra_pattern = (f'^(?:{cloud_pattern}{region_zone_pattern}|'
                     f'{wildcard_cloud}{wildcard_with_region}|'
                     f'{kubernetes_pattern})$')
    return infra_pattern


def _get_single_resources_schema():
    """Schema for a single resource in a resources list."""
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'cloud': {
                'type': 'string',
                'case_insensitive_enum': list(constants.ALL_CLOUDS)
            },
            'region': {
                'type': 'string',
            },
            'zone': {
                'type': 'string',
            },
            'infra': {
                'type': 'string',
                'description':
                    ('Infrastructure specification in format: '
                     'cloud[/region[/zone]]. Use "*" as a wildcard.'),
                # Pattern validates:
                # 1. cloud[/region[/zone]] - e.g. "aws", "aws/us-east-1",
                #    "aws/us-east-1/us-east-1a"
                # 2. Wildcard patterns - e.g. "*", "*/us-east-1",
                #    "*/*/us-east-1a", "aws/*/us-east-1a"
                # 3. Kubernetes patterns - e.g. "kubernetes/my-context",
                #    "k8s/context-name",
                #    "k8s/aws:eks:us-east-1:123456789012:cluster/my-cluster"
                'pattern': _get_infra_pattern(),
            },
            'cpus': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'number',
                }],
            },
            'memory': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'number',
                }],
            },
            'accelerators': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': [],
                    'maxProperties': 1,
                    'additionalProperties': {
                        'type': 'number'
                    }
                }]
            },
            'instance_type': {
                'type': 'string',
            },
            'use_spot': {
                'type': 'boolean',
            },
            'job_recovery': {
                # Either a string or a dict.
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': [],
                    'additionalProperties': False,
                    'properties': {
                        'strategy': {
                            'anyOf': [{
                                'type': 'string',
                            }, {
                                'type': 'null',
                            }],
                        },
                        'max_restarts_on_errors': {
                            'type': 'integer',
                            'minimum': 0,
                        },
                    }
                }],
            },
            'volumes': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'disk_size': {
                            'anyOf': [{
                                'type': 'string',
                                'pattern': constants.MEMORY_SIZE_PATTERN,
                            }, {
                                'type': 'integer',
                            }],
                        },
                        'disk_tier': {
                            'type': 'string',
                        },
                        'path': {
                            'type': 'string',
                        },
                        'auto_delete': {
                            'type': 'boolean',
                        },
                        'storage_type': {
                            'type': 'string',
                        },
                        'name': {
                            'type': 'string',
                        },
                        'attach_mode': {
                            'type': 'string',
                        },
                    },
                },
            },
            'disk_size': {
                'anyOf': [{
                    'type': 'string',
                    'pattern': constants.MEMORY_SIZE_PATTERN,
                }, {
                    'type': 'integer',
                }],
            },
            'disk_tier': {
                'type': 'string',
            },
            'network_tier': {
                'type': 'string',
            },
            'ports': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'integer',
                }, {
                    'type': 'array',
                    'items': {
                        'anyOf': [{
                            'type': 'string',
                        }, {
                            'type': 'integer',
                        }]
                    }
                }, {
                    'type': 'null',
                }],
            },
            'labels': {
                'type': 'object',
                'additionalProperties': {
                    'type': 'string'
                }
            },
            'accelerator_args': {
                'type': 'object',
                'required': [],
                'additionalProperties': False,
                'properties': {
                    'runtime_version': {
                        'type': 'string',
                    },
                    'tpu_name': {
                        'type': 'string',
                    },
                    'tpu_vm': {
                        'type': 'boolean',
                    }
                }
            },
            '_no_missing_accel_warnings': {
                'type': 'boolean'
            },
            'image_id': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': [],
                }, {
                    'type': 'null',
                }]
            },
            'autostop': _AUTOSTOP_SCHEMA,
            'priority': {
                'type': 'integer',
                'minimum': constants.MIN_PRIORITY,
                'maximum': constants.MAX_PRIORITY,
            },
            # The following fields are for internal use only. Should not be
            # specified in the task config.
            '_docker_login_config': {
                'type': 'object',
                'required': ['username', 'password', 'server'],
                'additionalProperties': False,
                'properties': {
                    'username': {
                        'type': 'string',
                    },
                    'password': {
                        'type': 'string',
                    },
                    'server': {
                        'type': 'string',
                    }
                }
            },
            '_is_image_managed': {
                'type': 'boolean',
            },
            '_requires_fuse': {
                'type': 'boolean',
            },
            '_cluster_config_overrides': {
                'type': 'object',
            },
        }
    }


def _get_multi_resources_schema():
    multi_resources_schema = {
        k: v
        for k, v in _get_single_resources_schema().items()
        # Validation may fail if $schema is included.
        if k != '$schema'
    }
    return multi_resources_schema


def get_resources_schema():
    """Resource schema in task config."""
    single_resources_schema = _get_single_resources_schema()['properties']
    single_resources_schema.pop('accelerators')
    multi_resources_schema = _get_multi_resources_schema()
    return {
        '$schema': 'http://json-schema.org/draft-07/schema#',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            **single_resources_schema,
            # We redefine the 'accelerators' field to allow one line list or
            # a set of accelerators.
            'accelerators': {
                # {'V100:1', 'A100:1'} will be
                # read as a string and converted to dict.
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': [],
                    'additionalProperties': {
                        'anyOf': [{
                            'type': 'null',
                        }, {
                            'type': 'number',
                        }]
                    }
                }, {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    }
                }]
            },
            'any_of': {
                'type': 'array',
                'items': multi_resources_schema,
            },
            'ordered': {
                'type': 'array',
                'items': multi_resources_schema,
            }
        },
    }


def get_volume_schema():
    # pylint: disable=import-outside-toplevel
    from sky.utils import volume

    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': ['name', 'type'],
        'additionalProperties': False,
        'properties': {
            'name': {
                'type': 'string',
            },
            'type': {
                'type': 'string',
                'case_sensitive_enum': [
                    type.value for type in volume.VolumeType
                ],
            },
            'infra': {
                'type': 'string',
                'description': ('Infrastructure specification in format: '
                                'cloud[/region[/zone]].'),
                # Pattern validates:
                # 1. cloud[/region[/zone]] - e.g. "aws", "aws/us-east-1",
                #    "aws/us-east-1/us-east-1a"
                # 2. Kubernetes patterns - e.g. "kubernetes/my-context",
                #    "k8s/context-name",
                #    "k8s/aws:eks:us-east-1:123456789012:cluster/my-cluster"
                'pattern': _get_volume_infra_pattern(),
            },
            'size': {
                'type': 'string',
                'pattern': constants.MEMORY_SIZE_PATTERN,
            },
            'resource_name': {
                'type': 'string',
            },
            'config': {
                'type': 'object',
                'required': [],
                'properties': {
                    'storage_class_name': {
                        'type': 'string',
                    },
                    'access_mode': {
                        'type': 'string',
                        'case_sensitive_enum': [
                            type.value for type in volume.VolumeAccessMode
                        ],
                    },
                    'namespace': {
                        'type': 'string',
                    },
                },
            },
            **_LABELS_SCHEMA,
        }
    }


def get_storage_schema():
    # pylint: disable=import-outside-toplevel
    from sky.data import storage

    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'name': {
                'type': 'string',
            },
            'source': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'array',
                    'minItems': 1,
                    'items': {
                        'type': 'string'
                    }
                }]
            },
            'store': {
                'type': 'string',
                'case_insensitive_enum': [
                    type.value for type in storage.StoreType
                ]
            },
            'persistent': {
                'type': 'boolean',
            },
            'mode': {
                'type': 'string',
                'case_insensitive_enum': [
                    mode.value for mode in storage.StorageMode
                ]
            },
            'config': {
                'type': 'object',
                'properties': {
                    'disk_size': {
                        'anyOf': [{
                            'type': 'string',
                            'pattern': constants.MEMORY_SIZE_PATTERN,
                        }, {
                            'type': 'integer',
                        }],
                    },
                    'disk_tier': {
                        'type': 'string',
                    },
                    'storage_type': {
                        'type': 'string',
                    },
                    'attach_mode': {
                        'type': 'string',
                    },
                },
            },
            '_is_sky_managed': {
                'type': 'boolean',
            },
            '_bucket_sub_path': {
                'type': 'string',
            },
            '_force_delete': {
                'type': 'boolean',
            }
        }
    }


def get_volume_mount_schema():
    """Schema for volume mount object in task config (internal use only)."""
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'path': {
                'type': 'string',
            },
            'volume_name': {
                'type': 'string',
            },
            'volume_config': {
                'type': 'object',
                'required': [],
                'additionalProperties': True,
                'properties': {
                    'cloud': {
                        'type': 'string',
                        'case_insensitive_enum': list(constants.ALL_CLOUDS)
                    },
                    'region': {
                        'anyOf': [{
                            'type': 'string'
                        }, {
                            'type': 'null'
                        }]
                    },
                    'zone': {
                        'anyOf': [{
                            'type': 'string'
                        }, {
                            'type': 'null'
                        }]
                    },
                },
            }
        }
    }


def get_service_schema():
    """Schema for top-level `service:` field (for SkyServe)."""
    # To avoid circular imports, only import when needed.
    # pylint: disable=import-outside-toplevel
    from sky.serve import load_balancing_policies
    from sky.serve import spot_placer
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            'readiness_probe': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': ['path'],
                    'additionalProperties': False,
                    'properties': {
                        'path': {
                            'type': 'string',
                        },
                        'initial_delay_seconds': {
                            'type': 'number',
                        },
                        'timeout_seconds': {
                            'type': 'number',
                        },
                        'post_data': {
                            'anyOf': [{
                                'type': 'string',
                            }, {
                                'type': 'object',
                            }]
                        },
                        'headers': {
                            'type': 'object',
                            'additionalProperties': {
                                'type': 'string'
                            }
                        },
                    }
                }]
            },
            'pool': {
                'type': 'boolean',
            },
            'replica_policy': {
                'type': 'object',
                'required': ['min_replicas'],
                'additionalProperties': False,
                'properties': {
                    'min_replicas': {
                        'type': 'integer',
                        'minimum': 0,
                    },
                    'max_replicas': {
                        'type': 'integer',
                        'minimum': 0,
                    },
                    'num_overprovision': {
                        'type': 'integer',
                        'minimum': 0,
                    },
                    'target_qps_per_replica': {
                        'anyOf': [
                            {
                                'type': 'number',
                                'minimum': 0,
                            },
                            {
                                'type': 'object',
                                'patternProperties': {
                                    # Pattern for accelerator types like
                                    # "H100:1", "A100:1", "H100", "A100"
                                    '^[A-Z0-9]+(?::[0-9]+)?$': {
                                        'type': 'number',
                                        'minimum': 0,
                                    }
                                },
                                'additionalProperties': False,
                            }
                        ]
                    },
                    'dynamic_ondemand_fallback': {
                        'type': 'boolean',
                    },
                    'base_ondemand_fallback_replicas': {
                        'type': 'integer',
                        'minimum': 0,
                    },
                    'spot_placer': {
                        'type': 'string',
                        'case_insensitive_enum': list(
                            spot_placer.SPOT_PLACERS.keys())
                    },
                    'upscale_delay_seconds': {
                        'type': 'number',
                    },
                    'downscale_delay_seconds': {
                        'type': 'number',
                    },
                }
            },
            'ports': {
                'type': 'integer',
            },
            'replicas': {
                'type': 'integer',
            },
            'workers': {
                'type': 'integer',
            },
            'load_balancing_policy': {
                'type': 'string',
                'case_insensitive_enum': list(
                    load_balancing_policies.LB_POLICIES.keys())
            },
            'tls': {
                'type': 'object',
                'required': ['keyfile', 'certfile'],
                'additionalProperties': False,
                'properties': {
                    'keyfile': {
                        'type': 'string',
                    },
                    'certfile': {
                        'type': 'string',
                    },
                },
            },
        }
    }


def _filter_schema(schema: dict, keys_to_keep: List[Tuple[str, ...]]) -> dict:
    """Recursively filter a schema to include only certain keys.

    Args:
        schema: The original schema dictionary.
        keys_to_keep: List of tuples with the path of keys to retain.

    Returns:
        The filtered schema.
    """
    # Convert list of tuples to a dictionary for easier access
    paths_dict: Dict[str, Any] = {}
    for path in keys_to_keep:
        current = paths_dict
        for step in path:
            if step not in current:
                current[step] = {}
            current = current[step]

    def keep_keys(current_schema: dict, current_path_dict: dict,
                  new_schema: dict) -> dict:
        # Base case: if we reach a leaf in the path_dict, we stop.
        if (not current_path_dict or not isinstance(current_schema, dict) or
                not current_schema.get('properties')):
            return current_schema

        if 'properties' not in new_schema:
            new_schema = {
                key: current_schema[key]
                for key in current_schema
                # We do not support the handling of `oneOf`, `anyOf`, `allOf`,
                # `required` for now.
                if key not in
                {'properties', 'oneOf', 'anyOf', 'allOf', 'required'}
            }
            new_schema['properties'] = {}
        for key, sub_schema in current_schema['properties'].items():
            if key in current_path_dict:
                # Recursively keep keys if further path dict exists
                new_schema['properties'][key] = {}
                current_path_value = current_path_dict.pop(key)
                new_schema['properties'][key] = keep_keys(
                    sub_schema, current_path_value,
                    new_schema['properties'][key])

        return new_schema

    # Start the recursive filtering
    new_schema = keep_keys(schema, paths_dict, {})
    assert not paths_dict, f'Unprocessed keys: {paths_dict}'
    return new_schema


def _experimental_task_schema() -> dict:
    # TODO: experimental.config_overrides has been deprecated in favor of the
    # top-level `config` field. Remove in v0.11.0.
    config_override_schema = _filter_schema(
        get_config_schema(), constants.OVERRIDEABLE_CONFIG_KEYS_IN_TASK)
    return {
        'experimental': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'config_overrides': config_override_schema,
            }
        }
    }


def get_task_schema():
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'name': {
                'type': 'string',
            },
            'workdir': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'object',
                    'required': ['url'],
                    'additionalProperties': False,
                    'properties': {
                        'url': {
                            'type': 'string',
                        },
                        'ref': {
                            'type': 'string',
                        },
                    },
                }],
            },
            'event_callback': {
                'type': 'string',
            },
            'num_nodes': {
                'type': 'integer',
            },
            # resources config is validated separately using RESOURCES_SCHEMA
            'resources': {
                'type': 'object',
            },
            # storage config is validated separately using STORAGE_SCHEMA
            'file_mounts': {
                'type': 'object',
            },
            # service config is validated separately using SERVICE_SCHEMA
            'service': {
                'type': 'object',
            },
            'pool': {
                'type': 'object',
            },
            'setup': {
                'type': 'string',
            },
            'run': {
                'type': 'string',
            },
            'envs': {
                'type': 'object',
                'required': [],
                'patternProperties': {
                    # Checks env keys are valid env var names.
                    '^[a-zA-Z_][a-zA-Z0-9_]*$': {
                        'type': ['string', 'null']
                    }
                },
                'additionalProperties': False,
            },
            'secrets': {
                'type': 'object',
                'required': [],
                'patternProperties': {
                    # Checks secret keys are valid env var names.
                    '^[a-zA-Z_][a-zA-Z0-9_]*$': {
                        'type': ['string', 'null']
                    }
                },
                'additionalProperties': False,
            },
            # inputs and outputs are experimental
            'inputs': {
                'type': 'object',
                'required': [],
                'maxProperties': 1,
                'additionalProperties': {
                    'type': 'number'
                }
            },
            'outputs': {
                'type': 'object',
                'required': [],
                'maxProperties': 1,
                'additionalProperties': {
                    'type': 'number'
                }
            },
            'file_mounts_mapping': {
                'type': 'object',
            },
            'config': _filter_schema(
                get_config_schema(),
                constants.OVERRIDEABLE_CONFIG_KEYS_IN_TASK),
            # volumes config is validated separately using get_volume_schema
            'volumes': {
                'type': 'object',
            },
            'volume_mounts': {
                'type': 'array',
                'items': get_volume_mount_schema(),
            },
            '_metadata': {
                'type': 'object',
            },
            **_experimental_task_schema(),
        }
    }


def get_cluster_schema():
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': ['cluster', 'auth'],
        'additionalProperties': False,
        'properties': {
            'cluster': {
                'type': 'object',
                'required': ['ips', 'name'],
                'additionalProperties': False,
                'properties': {
                    'ips': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        }
                    },
                    'name': {
                        'type': 'string',
                    },
                }
            },
            'auth': {
                'type': 'object',
                'required': ['ssh_user', 'ssh_private_key'],
                'additionalProperties': False,
                'properties': {
                    'ssh_user': {
                        'type': 'string',
                    },
                    'ssh_private_key': {
                        'type': 'string',
                    },
                }
            },
            'python': {
                'type': 'string',
            },
        }
    }


_NETWORK_CONFIG_SCHEMA = {
    'use_internal_ips': {
        'type': 'boolean',
    },
    'ssh_proxy_command': {
        'oneOf': [{
            'type': 'string',
        }, {
            'type': 'null',
        }, {
            'type': 'object',
            'required': [],
            'additionalProperties': {
                'anyOf': [
                    {
                        'type': 'string'
                    },
                    {
                        'type': 'null'
                    },
                ]
            }
        }]
    },
}

_LABELS_SCHEMA = {
    'labels': {
        'type': 'object',
        'required': [],
        'additionalProperties': {
            'type': 'string',
        },
    }
}

_PROPERTY_NAME_OR_CLUSTER_NAME_TO_PROPERTY = {
    'oneOf': [
        {
            'type': 'string'
        },
        {
            # A list of single-element dict to pretain the
            # order.
            # Example:
            #  property_name:
            #    - my-cluster1-*: my-property-1
            #    - my-cluster2-*: my-property-2
            #    - "*"": my-property-3
            'type': 'array',
            'items': {
                'type': 'object',
                'additionalProperties': {
                    'type': 'string'
                },
                'maxProperties': 1,
                'minProperties': 1,
            },
        }
    ]
}


class RemoteIdentityOptions(enum.Enum):
    """Enum for remote identity types.

    Some clouds (e.g., AWS, Kubernetes) also allow string values for remote
    identity, which map to the service account/role to use. Those are not
    included in this enum.
    """
    LOCAL_CREDENTIALS = 'LOCAL_CREDENTIALS'
    SERVICE_ACCOUNT = 'SERVICE_ACCOUNT'
    NO_UPLOAD = 'NO_UPLOAD'


def get_default_remote_identity(cloud: str) -> str:
    """Get the default remote identity for the specified cloud."""
    if cloud == 'kubernetes':
        return RemoteIdentityOptions.SERVICE_ACCOUNT.value
    return RemoteIdentityOptions.LOCAL_CREDENTIALS.value


_REMOTE_IDENTITY_SCHEMA = {
    'remote_identity': {
        'type': 'string',
        'case_insensitive_enum': [
            option.value for option in RemoteIdentityOptions
        ]
    }
}

_REMOTE_IDENTITY_SCHEMA_KUBERNETES = {
    'remote_identity': {
        'anyOf': [{
            'type': 'string'
        }, {
            'type': 'object',
            'additionalProperties': {
                'type': 'string'
            }
        }]
    },
}

_CONTEXT_CONFIG_SCHEMA_KUBERNETES = {
    # TODO(kevin): Remove 'networking' in v0.13.0.
    'networking': {
        'type': 'string',
        'case_insensitive_enum': [
            type.value for type in kubernetes_enums.KubernetesNetworkingMode
        ],
    },
    'ports': {
        'type': 'string',
        'case_insensitive_enum': [
            type.value for type in kubernetes_enums.KubernetesPortMode
        ],
    },
    'pod_config': {
        'type': 'object',
        'required': [],
        # Allow arbitrary keys since validating pod spec is hard
        'additionalProperties': True,
    },
    'custom_metadata': {
        'type': 'object',
        'required': [],
        # Allow arbitrary keys since validating metadata is hard
        'additionalProperties': True,
        # Disallow 'name' and 'namespace' keys in this dict
        'not': {
            'anyOf': [{
                'required': ['name']
            }, {
                'required': ['namespace']
            }]
        },
    },
    'provision_timeout': {
        'type': 'integer',
    },
    'autoscaler': {
        'type': 'string',
        'case_insensitive_enum': [
            type.value for type in kubernetes_enums.KubernetesAutoscalerType
        ],
    },
    'high_availability': {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'storage_class_name': {
                'type': 'string',
            }
        },
    },
    'kueue': {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'local_queue_name': {
                'type': 'string',
            },
        },
    },
    'dws': {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'enabled': {
                'type': 'boolean',
            },
            # Only used when Kueue is enabled.
            'max_run_duration': {
                'anyOf': [{
                    'type': 'string',
                    'pattern': constants.TIME_PATTERN,
                }, {
                    'type': 'integer',
                }]
            },
        },
    },
    'remote_identity': {
        'type': 'string',
    }
}


def get_config_schema():
    # pylint: disable=import-outside-toplevel
    from sky.server import daemons

    resources_schema = {
        k: v
        for k, v in get_resources_schema().items()
        # Validation may fail if $schema is included.
        if k != '$schema'
    }
    resources_schema['properties'].pop('ports')

    def _get_controller_schema():
        return {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'controller': {
                    'type': 'object',
                    'required': [],
                    'additionalProperties': False,
                    'properties': {
                        'resources': resources_schema,
                        'high_availability': {
                            'type': 'boolean',
                            'default': False,
                        },
                        'autostop': _AUTOSTOP_SCHEMA,
                        'consolidation_mode': {
                            'type': 'boolean',
                            'default': False,
                        }
                    },
                },
                'bucket': {
                    'type': 'string',
                    'pattern': '^(https|s3|gs|r2|cos)://.+',
                    'required': [],
                },
                'force_disable_cloud_bucket': {
                    'type': 'boolean',
                    'default': False,
                },
            }
        }

    cloud_configs = {
        'aws': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'prioritize_reservations': {
                    'type': 'boolean',
                },
                'specific_reservations': {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    },
                },
                'disk_encrypted': {
                    'type': 'boolean',
                },
                'ssh_user': {
                    'type': 'string',
                },
                'security_group_name':
                    (_PROPERTY_NAME_OR_CLUSTER_NAME_TO_PROPERTY),
                'vpc_name': {
                    'oneOf': [{
                        'type': 'string',
                    }, {
                        'type': 'null',
                    }],
                },
                'use_ssm': {
                    'type': 'boolean',
                },
                'post_provision_runcmd': {
                    'type': 'array',
                    'items': {
                        'oneOf': [{
                            'type': 'string'
                        }, {
                            'type': 'array',
                            'items': {
                                'type': 'string'
                            }
                        }]
                    },
                },
                **_LABELS_SCHEMA,
                **_NETWORK_CONFIG_SCHEMA,
            },
            **_check_not_both_fields_present('instance_tags', 'labels')
        },
        'gcp': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'prioritize_reservations': {
                    'type': 'boolean',
                },
                'specific_reservations': {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    },
                },
                'managed_instance_group': {
                    'type': 'object',
                    'required': ['run_duration'],
                    'additionalProperties': False,
                    'properties': {
                        'run_duration': {
                            'type': 'integer',
                        },
                        'provision_timeout': {
                            'type': 'integer',
                        }
                    }
                },
                'force_enable_external_ips': {
                    'type': 'boolean'
                },
                'enable_gvnic': {
                    'type': 'boolean'
                },
                'enable_gpu_direct': {
                    'type': 'boolean'
                },
                'placement_policy': {
                    'type': 'string',
                },
                'vpc_name': {
                    'oneOf': [
                        {
                            'type': 'string',
                            # vpc-name or project-id/vpc-name
                            # VPC name and Project ID have -, a-z, and 0-9.
                            'pattern': '^(?:[-a-z0-9]+/)?[-a-z0-9]+$'
                        },
                        {
                            'type': 'null',
                        }
                    ],
                },
                **_LABELS_SCHEMA,
                **_NETWORK_CONFIG_SCHEMA,
            },
            **_check_not_both_fields_present('instance_tags', 'labels')
        },
        'azure': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'storage_account': {
                    'type': 'string',
                },
                'resource_group_vm': {
                    'type': 'string',
                },
            }
        },
        'kubernetes': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'allowed_contexts': {
                    'oneOf': [{
                        'type': 'array',
                        'items': {
                            'type': 'string',
                        },
                    }, {
                        'type': 'string',
                        'pattern': '^all$'
                    }]
                },
                'context_configs': {
                    'type': 'object',
                    'required': [],
                    'properties': {},
                    # Properties are kubernetes context names.
                    'additionalProperties': {
                        'type': 'object',
                        'required': [],
                        'additionalProperties': False,
                        'properties': {
                            **_CONTEXT_CONFIG_SCHEMA_KUBERNETES,
                        },
                    },
                },
                **_CONTEXT_CONFIG_SCHEMA_KUBERNETES,
            }
        },
        'ssh': {
            'type': 'object',
            'required': [],
            'additionalProperties': False,
            'properties': {
                'allowed_node_pools': {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    },
                },
                'pod_config': {
                    'type': 'object',
                    'required': [],
                    # Allow arbitrary keys since validating pod spec is hard
                    'additionalProperties': True,
                },
            }
        },
        'oci': {
            'type': 'object',
            'required': [],
            'properties': {
                'region_configs': {
                    'type': 'object',
                    'required': [],
                    'properties': {},
                    # Properties are either 'default' or a region name.
                    'additionalProperties': {
                        'type': 'object',
                        'required': [],
                        'additionalProperties': False,
                        'properties': {
                            'compartment_ocid': {
                                'type': 'string',
                            },
                            'image_tag_general': {
                                'type': 'string',
                            },
                            'image_tag_gpu': {
                                'type': 'string',
                            },
                            'vcn_ocid': {
                                'type': 'string',
                            },
                            'vcn_subnet': {
                                'type': 'string',
                            },
                        }
                    },
                }
            },
        },
        'nebius': {
            'type': 'object',
            'required': [],
            'properties': {
                **_NETWORK_CONFIG_SCHEMA, 'use_static_ip_address': {
                    'type': 'boolean',
                },
                'tenant_id': {
                    'type': 'string',
                },
                'domain': {
                    'type': 'string',
                },
                'region_configs': {
                    'type': 'object',
                    'required': [],
                    'properties': {},
                    'additionalProperties': {
                        'type': 'object',
                        'required': [],
                        'additionalProperties': False,
                        'properties': {
                            'project_id': {
                                'type': 'string',
                            },
                            'fabric': {
                                'type': 'string',
                            },
                            'filesystems': {
                                'type': 'array',
                                'items': {
                                    'type': 'object',
                                    'additionalProperties': False,
                                    'properties': {
                                        'filesystem_id': {
                                            'type': 'string',
                                        },
                                        'attach_mode': {
                                            'type': 'string',
                                            'case_sensitive_enum': [
                                                'READ_WRITE', 'READ_ONLY'
                                            ]
                                        },
                                        'mount_path': {
                                            'type': 'string',
                                        }
                                    }
                                }
                            },
                        },
                    }
                }
            },
        }
    }

    admin_policy_schema = {
        'type': 'string',
        'anyOf': [
            {
                # Check regex to be a valid python module path
                'pattern': (r'^[a-zA-Z_][a-zA-Z0-9_]*'
                            r'(\.[a-zA-Z_][a-zA-Z0-9_]*)+$'),
            },
            {
                # Check for valid HTTP/HTTPS URL
                'pattern': r'^https?://.*$',
            }
        ]
    }

    allowed_clouds = {
        # A list of cloud names that are allowed to be used
        'type': 'array',
        'items': {
            'type': 'string',
            'case_insensitive_enum':
                (list(constants.ALL_CLOUDS) + ['cloudflare'])
        }
    }

    docker_configs = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'run_options': {
                'anyOf': [{
                    'type': 'string',
                }, {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    }
                }]
            }
        }
    }
    gpu_configs = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'disable_ecc': {
                'type': 'boolean',
            },
        }
    }

    daemon_config = {
        'type': 'object',
        'required': [],
        'properties': {
            'log_level': {
                'type': 'string',
                'case_insensitive_enum': ['DEBUG', 'INFO', 'WARNING'],
            },
        }
    }

    daemon_schema = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {}
    }

    for daemon in daemons.INTERNAL_REQUEST_DAEMONS:
        daemon_schema['properties'][daemon.id] = daemon_config

    api_server = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'endpoint': {
                'type': 'string',
                # Apply validation for URL
                'pattern': r'^https?://.*$',
            },
            'service_account_token': {
                'anyOf': [
                    {
                        'type': 'string',
                        # Validate that token starts with sky_ prefix
                        'pattern': r'^sky_.+$',
                    },
                    {
                        'type': 'null',
                    }
                ]
            },
            'requests_retention_hours': {
                'type': 'integer',
            },
            'cluster_event_retention_hours': {
                'type': 'number',
            },
            'cluster_debug_event_retention_hours': {
                'type': 'number',
            },
        }
    }

    rbac_schema = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'default_role': {
                'type': 'string',
                'case_insensitive_enum': ['admin', 'user']
            },
        },
    }

    workspace_schema = {'type': 'string'}

    allowed_workspace_cloud_names = list(constants.ALL_CLOUDS) + ['cloudflare']
    # Create pattern for not supported clouds, i.e.
    # all clouds except gcp, kubernetes, ssh
    not_supported_clouds = [
        cloud for cloud in allowed_workspace_cloud_names
        if cloud.lower() not in ['gcp', 'kubernetes', 'ssh', 'nebius']
    ]
    not_supported_cloud_regex = '|'.join(not_supported_clouds)
    workspaces_schema = {
        'type': 'object',
        'required': [],
        # each key is a workspace name
        'additionalProperties': {
            'type': 'object',
            'additionalProperties': False,
            'patternProperties': {
                # Pattern for non-GCP clouds - only allows 'disabled' property
                f'^({not_supported_cloud_regex})$': {
                    'type': 'object',
                    'additionalProperties': False,
                    'properties': {
                        'disabled': {
                            'type': 'boolean'
                        }
                    },
                },
            },
            'properties': {
                # Explicit definition for GCP allows both project_id and
                # disabled
                'private': {
                    'type': 'boolean',
                },
                'allowed_users': {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                    },
                },
                'gcp': {
                    'type': 'object',
                    'properties': {
                        'project_id': {
                            'type': 'string'
                        },
                        'disabled': {
                            'type': 'boolean'
                        }
                    },
                    'additionalProperties': False,
                },
                'ssh': {
                    'type': 'object',
                    'required': [],
                    'properties': {
                        'allowed_node_pools': {
                            'type': 'array',
                            'items': {
                                'type': 'string',
                            },
                        },
                        'disabled': {
                            'type': 'boolean'
                        },
                    },
                    'additionalProperties': False,
                },
                'kubernetes': {
                    'type': 'object',
                    'required': [],
                    'properties': {
                        'allowed_contexts': {
                            'oneOf': [{
                                'type': 'array',
                                'items': {
                                    'type': 'string',
                                },
                            }, {
                                'type': 'string',
                                'pattern': '^all$'
                            }]
                        },
                        'disabled': {
                            'type': 'boolean'
                        },
                    },
                    'additionalProperties': False,
                },
                'nebius': {
                    'type': 'object',
                    'required': [],
                    'properties': {
                        'credentials_file_path': {
                            'type': 'string',
                        },
                        'tenant_id': {
                            'type': 'string',
                        },
                        'domain': {
                            'type': 'string',
                        },
                        'disabled': {
                            'type': 'boolean'
                        },
                    },
                    'additionalProperties': False,
                },
            },
        },
    }

    provision_configs = {
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            'ssh_timeout': {
                'type': 'integer',
                'minimum': 1,
            },
        }
    }

    logs_schema = {
        'type': 'object',
        'required': ['store'],
        'additionalProperties': False,
        'properties': {
            'store': {
                'type': 'string',
                'case_insensitive_enum': ['gcp', 'aws'],
            },
            'gcp': {
                'type': 'object',
                'properties': {
                    'project_id': {
                        'type': 'string',
                    },
                    'credentials_file': {
                        'type': 'string',
                    },
                    'additional_labels': {
                        'type': 'object',
                        'additionalProperties': {
                            'type': 'string',
                        },
                    },
                },
            },
            'aws': {
                'type': 'object',
                'properties': {
                    'region': {
                        'type': 'string',
                    },
                    'credentials_file': {
                        'type': 'string',
                    },
                    'log_group_name': {
                        'type': 'string',
                    },
                    'log_stream_prefix': {
                        'type': 'string',
                    },
                    'auto_create_group': {
                        'type': 'boolean',
                    },
                    'additional_tags': {
                        'type': 'object',
                        'additionalProperties': {
                            'type': 'string',
                        },
                    },
                },
            },
        },
    }

    for cloud, config in cloud_configs.items():
        if cloud == 'aws':
            config['properties'].update(
                {'remote_identity': _PROPERTY_NAME_OR_CLUSTER_NAME_TO_PROPERTY})
        elif cloud == 'kubernetes':
            config['properties'].update(_REMOTE_IDENTITY_SCHEMA_KUBERNETES)
        else:
            config['properties'].update(_REMOTE_IDENTITY_SCHEMA)
    return {
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'type': 'object',
        'required': [],
        'additionalProperties': False,
        'properties': {
            # TODO Replace this with whatever syang cooks up
            'workspace': {
                'type': 'string',
            },
            'db': {
                'type': 'string',
            },
            'jobs': _get_controller_schema(),
            'serve': _get_controller_schema(),
            'allowed_clouds': allowed_clouds,
            'admin_policy': admin_policy_schema,
            'docker': docker_configs,
            'nvidia_gpus': gpu_configs,
            'api_server': api_server,
            'active_workspace': workspace_schema,
            'workspaces': workspaces_schema,
            'provision': provision_configs,
            'rbac': rbac_schema,
            'logs': logs_schema,
            'daemons': daemon_schema,
            **cloud_configs,
        },
    }



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ServiceNames(_message.Message):
    __slots__ = ("names",)
    NAMES_FIELD_NUMBER: _ClassVar[int]
    names: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, names: _Optional[_Iterable[str]] = ...) -> None: ...

class ServiceStatus(_message.Message):
    __slots__ = ("status",)
    class StatusEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: _containers.ScalarMap[str, str]
    def __init__(self, status: _Optional[_Mapping[str, str]] = ...) -> None: ...

class GetServiceStatusRequest(_message.Message):
    __slots__ = ("service_names", "pool")
    SERVICE_NAMES_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    service_names: ServiceNames
    pool: bool
    def __init__(self, service_names: _Optional[_Union[ServiceNames, _Mapping]] = ..., pool: bool = ...) -> None: ...

class GetServiceStatusResponse(_message.Message):
    __slots__ = ("statuses",)
    STATUSES_FIELD_NUMBER: _ClassVar[int]
    statuses: _containers.RepeatedCompositeFieldContainer[ServiceStatus]
    def __init__(self, statuses: _Optional[_Iterable[_Union[ServiceStatus, _Mapping]]] = ...) -> None: ...

class AddVersionRequest(_message.Message):
    __slots__ = ("service_name",)
    SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    service_name: str
    def __init__(self, service_name: _Optional[str] = ...) -> None: ...

class AddVersionResponse(_message.Message):
    __slots__ = ("version",)
    VERSION_FIELD_NUMBER: _ClassVar[int]
    version: int
    def __init__(self, version: _Optional[int] = ...) -> None: ...

class TerminateServicesRequest(_message.Message):
    __slots__ = ("service_names", "purge", "pool")
    SERVICE_NAMES_FIELD_NUMBER: _ClassVar[int]
    PURGE_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    service_names: ServiceNames
    purge: bool
    pool: bool
    def __init__(self, service_names: _Optional[_Union[ServiceNames, _Mapping]] = ..., purge: bool = ..., pool: bool = ...) -> None: ...

class TerminateServicesResponse(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class TerminateReplicaRequest(_message.Message):
    __slots__ = ("service_name", "replica_id", "purge")
    SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    REPLICA_ID_FIELD_NUMBER: _ClassVar[int]
    PURGE_FIELD_NUMBER: _ClassVar[int]
    service_name: str
    replica_id: int
    purge: bool
    def __init__(self, service_name: _Optional[str] = ..., replica_id: _Optional[int] = ..., purge: bool = ...) -> None: ...

class TerminateReplicaResponse(_message.Message):
    __slots__ = ("message",)
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class WaitServiceRegistrationRequest(_message.Message):
    __slots__ = ("service_name", "job_id", "pool")
    SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    service_name: str
    job_id: int
    pool: bool
    def __init__(self, service_name: _Optional[str] = ..., job_id: _Optional[int] = ..., pool: bool = ...) -> None: ...

class WaitServiceRegistrationResponse(_message.Message):
    __slots__ = ("lb_port",)
    LB_PORT_FIELD_NUMBER: _ClassVar[int]
    lb_port: int
    def __init__(self, lb_port: _Optional[int] = ...) -> None: ...

class UpdateServiceRequest(_message.Message):
    __slots__ = ("service_name", "version", "mode", "pool")
    SERVICE_NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    MODE_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    service_name: str
    version: int
    mode: str
    pool: bool
    def __init__(self, service_name: _Optional[str] = ..., version: _Optional[int] = ..., mode: _Optional[str] = ..., pool: bool = ...) -> None: ...

class UpdateServiceResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



"""Skylet configs."""
import functools
import os
import pathlib
import threading
from typing import Callable, Optional, Union

from sky.utils.db import db_utils

_DB_PATH = None
_db_init_lock = threading.Lock()


def init_db(func: Callable):
    """Ensure the table exists before calling the function.

    Since this module will be imported whenever `sky` is imported (due to
    Python's package importing logic), we should avoid creating the table
    until it's actually needed to avoid too many concurrent commit to the
    database.
    It solves the database locked problem in #1576.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _DB_PATH
        if _DB_PATH is not None:
            return func(*args, **kwargs)

        with _db_init_lock:
            if _DB_PATH is None:
                _DB_PATH = os.path.expanduser('~/.sky/skylet_config.db')
                os.makedirs(pathlib.Path(_DB_PATH).parents[0], exist_ok=True)
                with db_utils.safe_cursor(
                        _DB_PATH
                ) as c:  # Call it 'c' to avoid pylint complaining.
                    # Use WAL mode to avoid locking problem in #1507.
                    # Reference: https://stackoverflow.com/a/39265148
                    c.execute('PRAGMA journal_mode=WAL')
                    c.execute("""\
                        CREATE TABLE IF NOT EXISTS config (
                            key TEXT PRIMARY KEY,
                            value TEXT)""")
        return func(*args, **kwargs)

    return wrapper


@init_db
def get_config(key: str) -> Optional[bytes]:
    assert _DB_PATH is not None
    with db_utils.safe_cursor(_DB_PATH) as cursor:
        rows = cursor.execute('SELECT value FROM config WHERE key = ?', (key,))
        for (value,) in rows:
            return value
        return None


@init_db
def set_config(key: str, value: Union[bytes, str]) -> None:
    assert _DB_PATH is not None
    with db_utils.safe_cursor(_DB_PATH) as cursor:
        cursor.execute(
            """\
            INSERT OR REPLACE INTO config VALUES (?, ?)
            """, (key, value))



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AutostopWaitFor(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    AUTOSTOP_WAIT_FOR_UNSPECIFIED: _ClassVar[AutostopWaitFor]
    AUTOSTOP_WAIT_FOR_JOBS_AND_SSH: _ClassVar[AutostopWaitFor]
    AUTOSTOP_WAIT_FOR_JOBS: _ClassVar[AutostopWaitFor]
    AUTOSTOP_WAIT_FOR_NONE: _ClassVar[AutostopWaitFor]
AUTOSTOP_WAIT_FOR_UNSPECIFIED: AutostopWaitFor
AUTOSTOP_WAIT_FOR_JOBS_AND_SSH: AutostopWaitFor
AUTOSTOP_WAIT_FOR_JOBS: AutostopWaitFor
AUTOSTOP_WAIT_FOR_NONE: AutostopWaitFor

class SetAutostopRequest(_message.Message):
    __slots__ = ("idle_minutes", "backend", "wait_for", "down")
    IDLE_MINUTES_FIELD_NUMBER: _ClassVar[int]
    BACKEND_FIELD_NUMBER: _ClassVar[int]
    WAIT_FOR_FIELD_NUMBER: _ClassVar[int]
    DOWN_FIELD_NUMBER: _ClassVar[int]
    idle_minutes: int
    backend: str
    wait_for: AutostopWaitFor
    down: bool
    def __init__(self, idle_minutes: _Optional[int] = ..., backend: _Optional[str] = ..., wait_for: _Optional[_Union[AutostopWaitFor, str]] = ..., down: bool = ...) -> None: ...

class SetAutostopResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IsAutostoppingRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IsAutostoppingResponse(_message.Message):
    __slots__ = ("is_autostopping",)
    IS_AUTOSTOPPING_FIELD_NUMBER: _ClassVar[int]
    is_autostopping: bool
    def __init__(self, is_autostopping: bool = ...) -> None: ...



"""RunPod cloud adaptor."""

import os
import time
from typing import Any, Dict, Optional

from sky.adaptors import common

runpod = common.LazyImport(
    'runpod',
    import_error_message='Failed to import dependencies for RunPod. '
    'Try running: pip install "skypilot[runpod]"')

# Lazy imports
requests = common.LazyImport('requests')

_REST_BASE = 'https://rest.runpod.io/v1'
_MAX_RETRIES = 3
_TIMEOUT = 10


def _get_api_key() -> str:
    api_key = getattr(runpod, 'api_key', None)
    if not api_key:
        # Fallback to env if SDK global not set
        api_key = os.environ.get('RUNPOD_API_KEY')
    if not api_key:
        raise RuntimeError(
            'RunPod API key is not set. Please set runpod.api_key '
            'or RUNPOD_API_KEY.')
    return str(api_key)


def rest_request(method: str,
                 path: str,
                 json: Optional[Dict[str, Any]] = None) -> Any:
    url = f'{_REST_BASE}{path}'
    headers = {
        'Authorization': f'Bearer {_get_api_key()}',
        'Content-Type': 'application/json',
    }
    attempt = 0
    while True:
        attempt += 1
        try:
            resp = requests.request(method,
                                    url,
                                    headers=headers,
                                    json=json,
                                    timeout=_TIMEOUT)
        except Exception as e:  # pylint: disable=broad-except
            # Retry on transient network errors
            if attempt >= _MAX_RETRIES:
                raise RuntimeError(f'RunPod REST network error: {e}') from e
            time.sleep(1)
            continue

        # Retry on 5xx and 429
        if resp.status_code >= 500 or resp.status_code == 429:
            if attempt >= _MAX_RETRIES:
                raise RuntimeError(
                    f'RunPod REST error {resp.status_code}: {resp.text}')
            time.sleep(1)
            continue

        if resp.status_code >= 400:
            # Non-retryable client error
            raise RuntimeError(
                f'RunPod REST error {resp.status_code}: {resp.text}')

        if resp.text:
            try:
                return resp.json()
            except Exception:  # pylint: disable=broad-except
                return resp.text
        return None



"""Kubernetes pvc provisioning."""
from typing import Any, Dict, List, Optional, Set, Tuple

from sky import global_user_state
from sky import models
from sky import sky_logging
from sky.adaptors import kubernetes
from sky.provision.kubernetes import config as config_lib
from sky.provision.kubernetes import constants as k8s_constants
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import volume as volume_lib

logger = sky_logging.init_logger(__name__)


def _get_context_namespace(config: models.VolumeConfig) -> Tuple[str, str]:
    """Gets the context and namespace of a volume."""
    if config.region is None:
        context = kubernetes_utils.get_current_kube_config_context_name()
        config.region = context
    else:
        context = config.region
    namespace = config.config.get('namespace')
    if namespace is None:
        namespace = kubernetes_utils.get_kube_config_context_namespace(context)
        config.config['namespace'] = namespace
    return context, namespace


def check_pvc_usage_for_pod(context: Optional[str], namespace: str,
                            pod_spec: Dict[str, Any]) -> None:
    """Checks if the PVC is used by any pod in the namespace."""
    volumes = pod_spec.get('spec', {}).get('volumes', [])
    if not volumes:
        return
    once_modes = [
        volume_lib.VolumeAccessMode.READ_WRITE_ONCE.value,
        volume_lib.VolumeAccessMode.READ_WRITE_ONCE_POD.value
    ]
    for volume in volumes:
        pvc_name = volume.get('persistentVolumeClaim', {}).get('claimName')
        if not pvc_name:
            continue
        pvc = kubernetes.core_api(
            context).read_namespaced_persistent_volume_claim(
                name=pvc_name, namespace=namespace)
        access_mode = pvc.spec.access_modes[0]
        if access_mode not in once_modes:
            continue
        usedby_pods, _ = _get_volume_usedby(context, namespace, pvc_name)
        if usedby_pods:
            raise config_lib.KubernetesError(f'Volume {pvc_name} with access '
                                             f'mode {access_mode} is already '
                                             f'in use by Pods {usedby_pods}.')


def apply_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Creates or registers a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_spec = _get_pvc_spec(namespace, config)
    # Check if the storage class exists
    storage_class_name = pvc_spec['spec'].get('storageClassName')
    if storage_class_name is not None:
        try:
            kubernetes.storage_api(context).read_storage_class(
                name=storage_class_name)
        except kubernetes.api_exception() as e:
            raise config_lib.KubernetesError(
                f'Check storage class {storage_class_name} error: {e}')
    create_persistent_volume_claim(namespace, context, pvc_spec)
    return config


def delete_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Deletes a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud
    logger.info(f'Deleting PVC {pvc_name}')
    kubernetes_utils.delete_k8s_resource_with_retry(
        delete_func=lambda pvc_name=pvc_name: kubernetes.core_api(
            context).delete_namespaced_persistent_volume_claim(
                name=pvc_name,
                namespace=namespace,
                _request_timeout=config_lib.DELETION_TIMEOUT),
        resource_type='pvc',
        resource_name=pvc_name)
    return config


def _get_volume_usedby(
    context: Optional[str],
    namespace: str,
    pvc_name: str,
) -> Tuple[List[str], List[str]]:
    """Gets the usedby resources of a volume.

    This function returns the pods and clusters that are using the volume.
    The usedby_pods is accurate, which also includes the Pods that are not
    managed by SkyPilot.

    Args:
        context: Kubernetes context
        namespace: Kubernetes namespace
        pvc_name: PVC name

    Returns:
        usedby_pods: List of pods using the volume. These may include pods
                     not created by SkyPilot.
        usedby_clusters: List of clusters using the volume.
    """
    usedby_pods = []
    usedby_clusters = []
    field_selector = ','.join([
        f'status.phase!={phase}'
        for phase in k8s_constants.PVC_NOT_HOLD_POD_PHASES
    ])
    cloud_to_name_map = _get_cluster_name_on_cloud_to_cluster_name_map()
    # Get all pods in the namespace
    pods = kubernetes.core_api(context).list_namespaced_pod(
        namespace=namespace, field_selector=field_selector)
    for pod in pods.items:
        if pod.spec.volumes is None:
            continue
        for volume in pod.spec.volumes:
            if volume.persistent_volume_claim is None:
                continue
            if volume.persistent_volume_claim.claim_name == pvc_name:
                usedby_pods.append(pod.metadata.name)
                # Get the real cluster name
                cluster_name_on_cloud = pod.metadata.labels.get(
                    k8s_constants.TAG_SKYPILOT_CLUSTER_NAME)
                if cluster_name_on_cloud is None:
                    continue
                cluster_name = cloud_to_name_map.get(cluster_name_on_cloud)
                if cluster_name is not None:
                    usedby_clusters.append(cluster_name)
    if usedby_pods:
        logger.debug(f'Volume {pvc_name} is used by Pods {usedby_pods}'
                     f' and clusters {usedby_clusters}')
    return usedby_pods, usedby_clusters


def _get_cluster_name_on_cloud_to_cluster_name_map() -> Dict[str, str]:
    """Gets the map from cluster name on cloud to cluster name."""
    clusters = global_user_state.get_clusters()
    cloud_to_name_map = {}
    for cluster in clusters:
        handle = cluster['handle']
        if handle is None:
            continue
        cloud_to_name_map[handle.cluster_name_on_cloud] = cluster['name']
    return cloud_to_name_map


def get_volume_usedby(
    config: models.VolumeConfig,) -> Tuple[List[str], List[str]]:
    """Gets the usedby resources of a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud
    return _get_volume_usedby(context, namespace, pvc_name)


def get_all_volumes_usedby(
    configs: List[models.VolumeConfig],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Gets the usedby resources of all volumes."""
    field_selector = ','.join([
        f'status.phase!={phase}'
        for phase in k8s_constants.PVC_NOT_HOLD_POD_PHASES
    ])
    label_selector = 'parent=skypilot'
    context_to_namespaces: Dict[str, Set[str]] = {}
    pvc_names = set()
    for config in configs:
        context, namespace = _get_context_namespace(config)
        if context not in context_to_namespaces:
            context_to_namespaces[context] = set()
        context_to_namespaces[context].add(namespace)
        pvc_names.add(config.name_on_cloud)
    cloud_to_name_map = _get_cluster_name_on_cloud_to_cluster_name_map()
    # Get all pods in the namespace
    used_by_pods: Dict[str, Dict[str, Dict[str, List[str]]]] = {}
    used_by_clusters: Dict[str, Dict[str, Dict[str, List[str]]]] = {}
    for context, namespaces in context_to_namespaces.items():
        used_by_pods[context] = {}
        used_by_clusters[context] = {}
        for namespace in namespaces:
            used_by_pods[context][namespace] = {}
            used_by_clusters[context][namespace] = {}
            pods = kubernetes.core_api(context).list_namespaced_pod(
                namespace=namespace,
                field_selector=field_selector,
                label_selector=label_selector)
            for pod in pods.items:
                if pod.spec.volumes is None:
                    continue
                for volume in pod.spec.volumes:
                    if volume.persistent_volume_claim is None:
                        continue
                    volume_name = volume.persistent_volume_claim.claim_name
                    if volume_name not in pvc_names:
                        continue
                    if volume_name not in used_by_pods[context][namespace]:
                        used_by_pods[context][namespace][volume_name] = []
                    used_by_pods[context][namespace][volume_name].append(
                        pod.metadata.name)
                    cluster_name_on_cloud = pod.metadata.labels.get(
                        k8s_constants.TAG_SKYPILOT_CLUSTER_NAME)
                    if cluster_name_on_cloud is None:
                        continue
                    cluster_name = cloud_to_name_map.get(cluster_name_on_cloud)
                    if cluster_name is None:
                        continue
                    if cluster_name not in used_by_clusters[context][namespace]:
                        used_by_clusters[context][namespace][cluster_name] = []
                    used_by_clusters[context][namespace][cluster_name].append(
                        cluster_name)
    return used_by_pods, used_by_clusters


def map_all_volumes_usedby(
        used_by_pods: Dict[str, Any], used_by_clusters: Dict[str, Any],
        config: models.VolumeConfig) -> Tuple[List[str], List[str]]:
    """Maps the usedby resources of a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud

    return (used_by_pods.get(context, {}).get(namespace, {}).get(pvc_name, []),
            used_by_clusters.get(context, {}).get(namespace,
                                                  {}).get(pvc_name, []))


def create_persistent_volume_claim(namespace: str, context: Optional[str],
                                   pvc_spec: Dict[str, Any]) -> None:
    """Creates a persistent volume claim for SkyServe controller."""
    pvc_name = pvc_spec['metadata']['name']
    try:
        kubernetes.core_api(context).read_namespaced_persistent_volume_claim(
            name=pvc_name, namespace=namespace)
        logger.debug(f'PVC {pvc_name} already exists')
        return
    except kubernetes.api_exception() as e:
        if e.status != 404:  # Not found
            raise
    logger.info(f'Creating PVC {pvc_name}')
    kubernetes.core_api(context).create_namespaced_persistent_volume_claim(
        namespace=namespace, body=pvc_spec)


def _get_pvc_spec(namespace: str,
                  config: models.VolumeConfig) -> Dict[str, Any]:
    """Gets the PVC spec for the given storage config."""
    access_mode = config.config.get('access_mode')
    size = config.size
    # The previous code assumes that the access_mode and size are always set.
    assert access_mode is not None
    assert size is not None
    pvc_spec: Dict[str, Any] = {
        'metadata': {
            'name': config.name_on_cloud,
            'namespace': namespace,
            'labels': {
                'parent': 'skypilot',
                'skypilot-name': config.name,
            }
        },
        'spec': {
            'accessModes': [access_mode],
            'resources': {
                'requests': {
                    'storage': f'{size}Gi'
                }
            },
        }
    }
    if config.labels:
        pvc_spec['metadata']['labels'].update(config.labels)
    storage_class = config.config.get('storage_class_name')
    if storage_class is not None:
        pvc_spec['spec']['storageClassName'] = storage_class
    return pvc_spec



"""Utilities for streaming logs from response."""

import asyncio
import collections
import pathlib
from typing import AsyncGenerator, Deque, List, Optional

import aiofiles
import fastapi

from sky import global_user_state
from sky import sky_logging
from sky.server.requests import requests as requests_lib
from sky.utils import common_utils
from sky.utils import message_utils
from sky.utils import rich_utils
from sky.utils import status_lib

logger = sky_logging.init_logger(__name__)

# When streaming log lines, buffer the lines in memory and flush them in chunks
# to improve log tailing throughput. Buffer size is the max size bytes of each
# chunk and the timeout threshold for flushing the buffer to ensure
# responsiveness.
_BUFFER_SIZE = 8 * 1024  # 8KB
_BUFFER_TIMEOUT = 0.02  # 20ms
_HEARTBEAT_INTERVAL = 30
# If a SHORT request has been stuck in pending for
# _SHORT_REQUEST_SPINNER_TIMEOUT seconds, we show the waiting spinner
_SHORT_REQUEST_SPINNER_TIMEOUT = 2
# If there is an issue during provisioning that causes the cluster to be stuck
# in INIT state, we use this timeout to break the loop and stop streaming
# provision logs.
_PROVISION_LOG_TIMEOUT = 3
# Maximum time to wait for new log files to appear when streaming worker node
# provision logs. Worker logs are created sequentially during the provisioning
# process, so we need to wait for new files to appear.
_MAX_WAIT_FOR_NEW_LOG_FILES = 3  # seconds

LONG_REQUEST_POLL_INTERVAL = 1
DEFAULT_POLL_INTERVAL = 0.1


async def _yield_log_file_with_payloads_skipped(
        log_file) -> AsyncGenerator[str, None]:
    async for line in log_file:
        if not line:
            return
        is_payload, line_str = message_utils.decode_payload(
            line.decode('utf-8'), raise_for_mismatch=False)
        if is_payload:
            continue

        yield line_str


async def log_streamer(
    request_id: Optional[str],
    log_path: Optional[pathlib.Path] = None,
    plain_logs: bool = False,
    tail: Optional[int] = None,
    follow: bool = True,
    cluster_name: Optional[str] = None,
    polling_interval: float = DEFAULT_POLL_INTERVAL
) -> AsyncGenerator[str, None]:
    """Streams the logs of a request.

    Args:
        request_id: The request ID to check whether the log tailing process
            should be stopped.
        log_path: The path to the log file or directory containing the log
        files. If it is a directory, all *.log files in the directory will be
        streamed.
        plain_logs: Whether to show plain logs.
        tail: The number of lines to tail. If None, tail the whole file.
        follow: Whether to follow the log file.
        cluster_name: The cluster name to check status for provision logs.
            If provided and cluster status is UP, streaming will terminate.
    """

    if request_id is not None:
        start_time = asyncio.get_event_loop().time()
        status_msg = rich_utils.EncodedStatusMessage(
            f'[dim]Checking request: {request_id}[/dim]')
        request_task = await requests_lib.get_request_async(request_id,
                                                            fields=[
                                                                'request_id',
                                                                'name',
                                                                'schedule_type',
                                                                'status',
                                                                'status_msg'
                                                            ])

        if request_task is None:
            raise fastapi.HTTPException(
                status_code=404, detail=f'Request {request_id} not found')
        request_id = request_task.request_id

        # By default, do not show the waiting spinner for SHORT requests.
        # If the request has been stuck in pending for
        # _SHORT_REQUEST_SPINNER_TIMEOUT seconds, we show the waiting spinner
        show_request_waiting_spinner = (not plain_logs and
                                        request_task.schedule_type
                                        == requests_lib.ScheduleType.LONG)

        if show_request_waiting_spinner:
            yield status_msg.init()
            yield status_msg.start()
        last_waiting_msg = ''
        waiting_msg = (f'Waiting for {request_task.name!r} request to be '
                       f'scheduled: {request_id}')
        req_status = request_task.status
        req_msg = request_task.status_msg
        del request_task
        # Slowly back off the database polling up to every 1 second, to avoid
        # overloading the CPU and DB.
        backoff = common_utils.Backoff(initial_backoff=polling_interval,
                                       max_backoff_factor=10,
                                       multiplier=1.2)
        while req_status < requests_lib.RequestStatus.RUNNING:
            current_time = asyncio.get_event_loop().time()
            # Show the waiting spinner for a SHORT request if it has been stuck
            # in pending for _SHORT_REQUEST_SPINNER_TIMEOUT seconds
            if not show_request_waiting_spinner and (
                    current_time - start_time > _SHORT_REQUEST_SPINNER_TIMEOUT):
                show_request_waiting_spinner = True
                yield status_msg.init()
                yield status_msg.start()
            if req_msg is not None:
                waiting_msg = req_msg
            if show_request_waiting_spinner:
                yield status_msg.update(f'[dim]{waiting_msg}[/dim]')
            elif plain_logs and waiting_msg != last_waiting_msg:
                # Only log when waiting message changes.
                last_waiting_msg = waiting_msg
                # Use smaller padding (1024 bytes) to force browser rendering
                yield f'{waiting_msg}' + ' ' * 4096 + '\n'
            # Sleep shortly to avoid storming the DB and CPU and allow other
            # coroutines to run.
            # TODO(aylei): we should use a better mechanism to avoid busy
            # polling the DB, which can be a bottleneck for high-concurrency
            # requests.
            await asyncio.sleep(backoff.current_backoff())
            status_with_msg = await requests_lib.get_request_status_async(
                request_id, include_msg=True)
            req_status = status_with_msg.status
            req_msg = status_with_msg.status_msg
            if not follow:
                break
        if show_request_waiting_spinner:
            yield status_msg.stop()

    if log_path is not None and log_path.is_dir():
        # Track which log files we've already streamed
        streamed_files = set()
        no_new_files_count = 0

        while True:
            # Get all *.log files in the log_path
            log_files = sorted(log_path.glob('*.log'))

            # Filter out already streamed files
            new_files = [f for f in log_files if f not in streamed_files]

            if len(new_files) == 0:
                if not follow:
                    break
                # Wait a bit to see if new files appear
                await asyncio.sleep(0.5)
                no_new_files_count += 1
                # Check if we've waited too long for new files
                if no_new_files_count > _MAX_WAIT_FOR_NEW_LOG_FILES * 2:
                    break
                continue

            # Reset the no-new-files counter when we find new files
            no_new_files_count = 0

            for log_file_path in new_files:
                # Add header before each file (similar to tail -f behavior)
                header = f'\n==> {log_file_path} <==\n\n'
                yield header

                async with aiofiles.open(log_file_path, 'rb') as f:
                    async for chunk in _tail_log_file(f, request_id, plain_logs,
                                                      tail, follow,
                                                      cluster_name,
                                                      polling_interval):
                        yield chunk

                # Mark this file as streamed
                streamed_files.add(log_file_path)

            # If not following, break after streaming all current files
            if not follow:
                break
    else:
        assert log_path is not None, (request_id, log_path)
        async with aiofiles.open(log_path, 'rb') as f:
            async for chunk in _tail_log_file(f, request_id, plain_logs, tail,
                                              follow, cluster_name,
                                              polling_interval):
                yield chunk


async def _tail_log_file(
    f: aiofiles.threadpool.binary.AsyncBufferedReader,
    request_id: Optional[str] = None,
    plain_logs: bool = False,
    tail: Optional[int] = None,
    follow: bool = True,
    cluster_name: Optional[str] = None,
    polling_interval: float = DEFAULT_POLL_INTERVAL
) -> AsyncGenerator[str, None]:
    """Tail the opened log file, buffer the lines and flush in chunks."""

    if tail is not None:
        # Find last n lines of the log file. Do not read the whole file into
        # memory.
        # TODO(zhwu): this will include the control lines for rich status,
        # which may not lead to exact tail lines when showing on the client
        # side.
        lines: Deque[str] = collections.deque(maxlen=tail)
        async for line_str in _yield_log_file_with_payloads_skipped(f):
            lines.append(line_str)
        for line_str in lines:
            yield line_str

    last_heartbeat_time = asyncio.get_event_loop().time()
    last_status_check_time = asyncio.get_event_loop().time()

    # Buffer the lines in memory and flush them in chunks to improve log
    # tailing throughput.
    buffer: List[str] = []
    buffer_bytes = 0
    last_flush_time = asyncio.get_event_loop().time()

    async def flush_buffer() -> AsyncGenerator[str, None]:
        nonlocal buffer, buffer_bytes, last_flush_time
        if buffer:
            yield ''.join(buffer)
            buffer.clear()
            buffer_bytes = 0
            last_flush_time = asyncio.get_event_loop().time()

    while True:
        # Sleep 0 to yield control to allow other coroutines to run,
        # while keeps the loop tight to make log stream responsive.
        await asyncio.sleep(0)
        current_time = asyncio.get_event_loop().time()
        # Flush the buffer when it is not empty and the buffer is full or the
        # flush timeout is reached.
        if buffer and (buffer_bytes >= _BUFFER_SIZE or
                       (current_time - last_flush_time) >= _BUFFER_TIMEOUT):
            async for chunk in flush_buffer():
                yield chunk

        line: Optional[bytes] = await f.readline()
        if not line:
            # Avoid checking the status too frequently to avoid overloading the
            # DB.
            should_check_status = (current_time -
                                   last_status_check_time) >= polling_interval
            if not follow:
                # We will only hit this path once, but we should make sure to
                # check the status so that we display the final request status
                # if the request is complete.
                should_check_status = True
            if request_id is not None and should_check_status:
                last_status_check_time = current_time
                req_status = await requests_lib.get_request_status_async(
                    request_id)
                if req_status.status > requests_lib.RequestStatus.RUNNING:
                    if (req_status.status ==
                            requests_lib.RequestStatus.CANCELLED):
                        request_task = await requests_lib.get_request_async(
                            request_id, fields=['name', 'should_retry'])
                        if request_task.should_retry:
                            buffer.append(
                                message_utils.encode_payload(
                                    rich_utils.Control.RETRY.encode('')))
                        else:
                            buffer.append(
                                f'{request_task.name!r} request {request_id}'
                                ' cancelled\n')
                        del request_task
                    break
            if not follow:
                # The below checks (cluster status, heartbeat) are not needed
                # for non-follow logs.
                break
            # Provision logs pass in cluster_name, check cluster status
            # periodically to see if provisioning is done.
            if cluster_name is not None:
                if current_time - last_flush_time > _PROVISION_LOG_TIMEOUT:
                    break
                if should_check_status:
                    last_status_check_time = current_time
                    cluster_status = await (
                        global_user_state.get_status_from_cluster_name_async(
                            cluster_name))
                    if cluster_status is None:
                        logger.debug(
                            'Stop tailing provision logs for cluster'
                            f' status for cluster {cluster_name} not found')
                        break
                    if cluster_status != status_lib.ClusterStatus.INIT:
                        logger.debug(
                            f'Stop tailing provision logs for cluster'
                            f' {cluster_name} has status {cluster_status} '
                            '(not in INIT state)')
            if current_time - last_heartbeat_time >= _HEARTBEAT_INTERVAL:
                # Currently just used to keep the connection busy, refer to
                # https://github.com/skypilot-org/skypilot/issues/5750 for
                # more details.
                buffer.append(
                    message_utils.encode_payload(
                        rich_utils.Control.HEARTBEAT.encode('')))
                last_heartbeat_time = current_time

            # Sleep shortly to avoid storming the DB and CPU, this has
            # little impact on the responsivness here since we are waiting
            # for a new line to come in.
            await asyncio.sleep(0.1)
            continue

        # Refresh the heartbeat time, this is a trivial optimization for
        # performance but it helps avoid unnecessary heartbeat strings
        # being printed when the client runs in an old version.
        last_heartbeat_time = asyncio.get_event_loop().time()
        line_str = line.decode('utf-8')
        if plain_logs:
            is_payload, line_str = message_utils.decode_payload(
                line_str, raise_for_mismatch=False)
            # TODO(aylei): implement heartbeat mechanism for plain logs,
            # sending invisible characters might be okay.
            if is_payload:
                continue
        buffer.append(line_str)
        buffer_bytes += len(line_str.encode('utf-8'))

    # Flush remaining lines in the buffer.
    async for chunk in flush_buffer():
        yield chunk


def stream_response_for_long_request(
    request_id: str,
    logs_path: pathlib.Path,
    background_tasks: fastapi.BackgroundTasks,
    kill_request_on_disconnect: bool = True,
) -> fastapi.responses.StreamingResponse:
    """Stream the logs of a long request."""
    return stream_response(
        request_id,
        logs_path,
        background_tasks,
        polling_interval=LONG_REQUEST_POLL_INTERVAL,
        kill_request_on_disconnect=kill_request_on_disconnect,
    )


def stream_response(
    request_id: str,
    logs_path: pathlib.Path,
    background_tasks: fastapi.BackgroundTasks,
    polling_interval: float = DEFAULT_POLL_INTERVAL,
    kill_request_on_disconnect: bool = True,
) -> fastapi.responses.StreamingResponse:

    if kill_request_on_disconnect:

        async def on_disconnect():
            logger.info(f'User terminated the connection for request '
                        f'{request_id}')
            requests_lib.kill_requests([request_id])

        # The background task will be run after returning a response.
        # https://fastapi.tiangolo.com/tutorial/background-tasks/
        background_tasks.add_task(on_disconnect)

    return fastapi.responses.StreamingResponse(
        log_streamer(request_id, logs_path, polling_interval=polling_interval),
        media_type='text/plain',
        headers={
            'Cache-Control': 'no-cache, no-transform',
            'X-Accel-Buffering': 'no',
            'Transfer-Encoding': 'chunked'
        })



"""Cudo catalog helper."""

cudo_gpu_model = {
    'H100 NVL': 'H100',
    'H100 SXM': 'H100-SXM',
    'L40S (compute mode)': 'L40S',
    'L40S (graphics mode)': 'L40S',
    'A40 (compute mode)': 'A40',
    'A40 (graphics mode)': 'A40',
    'RTX A5000': 'RTXA5000',
    'RTX A6000': 'RTXA6000',
    'A100 80GB PCIe': 'A100',
    'A800 PCIe': 'A800',
    'V100': 'V100',
}

cudo_gpu_mem = {
    'H100': 94,
    'H100-SXM': 80,
    'L40S': 48,
    'A40': 48,
    'RTXA5000': 24,
    'RTXA6000': 48,
    'A100': 80,
    'A800': 80,
    'V100': 16,
}

machine_specs = [
    # Low
    {
        'vcpu': 2,
        'mem': 4,
        'gpu': 1,
    },
    {
        'vcpu': 4,
        'mem': 8,
        'gpu': 1,
    },
    {
        'vcpu': 8,
        'mem': 16,
        'gpu': 2,
    },
    {
        'vcpu': 16,
        'mem': 32,
        'gpu': 2,
    },
    {
        'vcpu': 32,
        'mem': 64,
        'gpu': 4,
    },
    {
        'vcpu': 64,
        'mem': 128,
        'gpu': 8,
    },
    # Mid
    {
        'vcpu': 96,
        'mem': 192,
        'gpu': 8
    },
    {
        'vcpu': 48,
        'mem': 96,
        'gpu': 4
    },
    {
        'vcpu': 24,
        'mem': 48,
        'gpu': 2
    },
    {
        'vcpu': 12,
        'mem': 24,
        'gpu': 1
    },
    # Hi
    {
        'vcpu': 96,
        'mem': 192,
        'gpu': 4
    },
    {
        'vcpu': 48,
        'mem': 96,
        'gpu': 2
    },
    {
        'vcpu': 24,
        'mem': 48,
        'gpu': 1
    },
]


def cudo_gpu_to_skypilot_gpu(model):
    if model in cudo_gpu_model:
        return cudo_gpu_model[model]
    else:
        return model


def skypilot_gpu_to_cudo_gpu(model):
    for key, value in cudo_gpu_model.items():
        if value == model:
            return key
    return model


def gpu_exists(model):
    if model in cudo_gpu_model:
        return True
    return False



"""SSH Node Pool management API endpoints."""
import re
from typing import Any, Dict, List

import fastapi

from sky import core as sky_core
from sky.server.requests import executor
from sky.server.requests import payloads
from sky.server.requests import requests as requests_lib
from sky.ssh_node_pools import core as ssh_node_pools_core
from sky.utils import common_utils

router = fastapi.APIRouter()


@router.get('')
def get_ssh_node_pools() -> Dict[str, Any]:
    """Get all SSH Node Pool configurations."""
    try:
        return ssh_node_pools_core.get_all_pools()
    except Exception as e:
        raise fastapi.HTTPException(
            status_code=500,
            detail=
            f'Failed to get SSH Node Pools: {common_utils.format_exception(e)}')


@router.post('')
def update_ssh_node_pools(pools_config: Dict[str, Any]) -> Dict[str, str]:
    """Update SSH Node Pool configurations."""
    try:
        ssh_node_pools_core.update_pools(pools_config)
        return {'status': 'success'}
    except Exception as e:
        raise fastapi.HTTPException(status_code=400,
                                    detail=f'Failed to update SSH Node Pools:'
                                    f' {common_utils.format_exception(e)}')


@router.delete('/{pool_name}')
def delete_ssh_node_pool(pool_name: str) -> Dict[str, str]:
    """Delete a SSH Node Pool configuration."""
    try:
        if ssh_node_pools_core.delete_pool(pool_name):
            return {'status': 'success'}
        else:
            raise fastapi.HTTPException(
                status_code=404,
                detail=f'SSH Node Pool `{pool_name}` not found')
    except fastapi.HTTPException:
        raise
    except Exception as e:
        raise fastapi.HTTPException(status_code=500,
                                    detail='Failed to delete SSH Node Pool: '
                                    f'{common_utils.format_exception(e)}')


@router.post('/keys')
async def upload_ssh_key(request: fastapi.Request) -> Dict[str, str]:
    """Upload SSH private key."""
    try:
        form = await request.form()
        key_name = form.get('key_name')
        key_file = form.get('key_file')

        if not key_name or not key_file:
            raise fastapi.HTTPException(status_code=400,
                                        detail='Missing key_name or key_file')

        key_content = await key_file.read()
        key_path = ssh_node_pools_core.upload_ssh_key(key_name,
                                                      key_content.decode())

        return {'status': 'success', 'key_path': key_path}
    except fastapi.HTTPException:
        raise
    except Exception as e:
        raise fastapi.HTTPException(
            status_code=500,
            detail=
            f'Failed to upload SSH key: {common_utils.format_exception(e)}')


@router.get('/keys')
def list_ssh_keys() -> List[str]:
    """List available SSH keys."""
    try:
        return ssh_node_pools_core.list_ssh_keys()
    except Exception as e:
        exception_msg = common_utils.format_exception(e)
        raise fastapi.HTTPException(
            status_code=500, detail=f'Failed to list SSH keys: {exception_msg}')


@router.post('/{pool_name}/deploy')
async def deploy_ssh_node_pool(request: fastapi.Request,
                               pool_name: str) -> Dict[str, str]:
    """Deploy SSH Node Pool using existing ssh_up functionality."""
    try:
        ssh_up_body = payloads.SSHUpBody(infra=pool_name, cleanup=False)
        await executor.schedule_request_async(
            request_id=request.state.request_id,
            request_name='ssh_up',
            request_body=ssh_up_body,
            func=sky_core.ssh_up,
            schedule_type=requests_lib.ScheduleType.LONG,
        )

        return {
            'status': 'success',
            'request_id': request.state.request_id,
            'message': f'SSH Node Pool `{pool_name}` deployment started'
        }
    except Exception as e:
        raise fastapi.HTTPException(status_code=500,
                                    detail=f'Failed to deploy SSH Node Pool: '
                                    f'{common_utils.format_exception(e)}')


@router.post('/deploy')
async def deploy_ssh_node_pool_general(
        request: fastapi.Request,
        ssh_up_body: payloads.SSHUpBody) -> Dict[str, str]:
    """Deploys all SSH Node Pools."""
    try:
        await executor.schedule_request_async(
            request_id=request.state.request_id,
            request_name='ssh_up',
            request_body=ssh_up_body,
            func=sky_core.ssh_up,
            schedule_type=requests_lib.ScheduleType.LONG,
        )

        pool_name = ssh_up_body.infra or 'default'
        return {
            'status': 'success',
            'request_id': request.state.request_id,
            'message': f'SSH Node Pool `{pool_name}` deployment started'
        }
    except Exception as e:
        raise fastapi.HTTPException(status_code=500,
                                    detail=f'Failed to deploy SSH Node Pool: '
                                    f'{common_utils.format_exception(e)}')


@router.post('/{pool_name}/down')
async def down_ssh_node_pool(request: fastapi.Request,
                             pool_name: str) -> Dict[str, str]:
    """Cleans up a SSH Node Pools."""
    try:
        ssh_up_body = payloads.SSHUpBody(infra=pool_name, cleanup=True)
        await executor.schedule_request_async(
            request_id=request.state.request_id,
            request_name='ssh_down',
            request_body=ssh_up_body,
            func=sky_core.ssh_up,  # Reuse ssh_up function with cleanup=True
            schedule_type=requests_lib.ScheduleType.LONG,
        )

        return {
            'status': 'success',
            'request_id': request.state.request_id,
            'message': f'SSH Node Pool `{pool_name}` teardown started'
        }
    except Exception as e:
        raise fastapi.HTTPException(
            status_code=500,
            detail=f'Failed to tear down SSH Node Pool: '
            f'{common_utils.format_exception(e)}')


@router.post('/down')
async def down_ssh_node_pool_general(
        request: fastapi.Request,
        ssh_up_body: payloads.SSHUpBody) -> Dict[str, str]:
    """Cleans up all SSH Node Pools."""
    try:
        # Set cleanup=True for down operation
        ssh_up_body.cleanup = True
        await executor.schedule_request_async(
            request_id=request.state.request_id,
            request_name='ssh_down',
            request_body=ssh_up_body,
            func=sky_core.ssh_up,  # Reuse ssh_up function with cleanup=True
            schedule_type=requests_lib.ScheduleType.LONG,
        )

        pool_name = ssh_up_body.infra or 'default'
        return {
            'status': 'success',
            'request_id': request.state.request_id,
            'message': f'SSH Node Pool `{pool_name}` teardown started'
        }
    except Exception as e:
        raise fastapi.HTTPException(
            status_code=500,
            detail=f'Failed to tear down SSH Node Pool: '
            f'{common_utils.format_exception(e)}')


@router.get('/{pool_name}/status')
def get_ssh_node_pool_status(pool_name: str) -> Dict[str, str]:
    """Get the status of a specific SSH Node Pool."""
    try:
        # Call ssh_status to check the context
        context_name = f'ssh-{pool_name}'
        is_ready, reason = sky_core.ssh_status(context_name)

        # Strip ANSI escape codes from the reason text
        def strip_ansi_codes(text):
            if not text:
                return text
            # Remove ANSI escape sequences (color codes, formatting, etc.)
            text = re.sub(r'\x1b\[[0-9;]*m', '', text)
            # Remove 'disabled. Reason: ' prefix if present
            text = text.replace('disabled. Reason: ', '')
            return text

        cleaned_reason = strip_ansi_codes(reason) if reason else reason

        return {
            'pool_name': pool_name,
            'context_name': context_name,
            'status': 'Ready' if is_ready else 'Not Ready',
            'reason': cleaned_reason
        }
    except Exception as e:
        raise fastapi.HTTPException(
            status_code=500,
            detail=f'Failed to get SSH Node Pool status: '
            f'{common_utils.format_exception(e)}')
